#!/usr/bin/env python
# -*- coding: utf-8 -*-

#Copyright (c) 2015 3D Control Systems LTD

#3DPrinterOS client is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#3DPrinterOS client is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with 3DPrinterOS client.  If not, see <http://www.gnu.org/licenses/>.

# Author: Vladimir Avdeev <another.vic@yandex.ru>

import time
import signal
import platform
import threading

import log
import paths
paths.init_path_to_libs()
import usb_detect
import http_client
import tray_controller
import camera_controller
import cloud_sync_controller
import user_login
import updater
import version
import printer_interface
import config
import rights
import makerware_utils
import network_detect


class App(object):

    MAIN_LOOP_SLEEP = 1
    IDLE_RECORD_TIMEOUT = 120
    VIRTUAL_PRINTER_USB_INFO = {"VID" : "ZZZZ", "PID": "ZZZZ", "SNR": "0", "COM": None}
    BROWSER_OPENING_DELAY = 2

    @log.log_exception
    def __init__(self):
        config.Config.instance().set_app_pointer(self)
        self.logger = log.create_logger('', log.LOG_FILE)
        self.logger.info("Starting 3DPrinterOS client. Version: %s%s%s" %
                         (version.version, version.branch, version.build))
        self.logger.info('Operating system: ' + platform.system() + ' ' + platform.release())
        signal.signal(signal.SIGINT, self.intercept_signal)
        signal.signal(signal.SIGTERM, self.intercept_signal)
        self.printer_interfaces = []
        self.virtual_printer_enabled = False
        self.stop_flag = False
        self.restart_flag = False
        self.closing_status = []
        self.detector = usb_detect.USBDetector()
        self.network_printers_detector = network_detect.NetworkDetector(self)
        self.rights_checker_waiter = rights.RightsCheckerWaiter(self)
        self.conveyor_kill_waiter = makerware_utils.ConveyorKillWaiter(self)
        self.network_connection_checker = http_client.NetworkConnectionChecker(self)
        self.user_login = user_login.UserLogin(self)
        self.init_interface()
        threading.Thread(target=self.wait_and_open_browser()).start()
        self.network_connection_checker.wait()
        self.rights_checker_waiter.wait()
        self.conveyor_kill_waiter.wait()
        self.user_login.wait()
        if self.user_login.user_token and hasattr(self.user_login, "profiles"):
            config.Config.instance().set_profiles(self.user_login.profiles)
            self.updater = updater.Updater()
            self.virtual_printer_enabled = config.get_settings()['virtual_printer']['enabled']
            self.tray_controller = tray_controller.TrayController()
            self.camera_controller = camera_controller.CameraController(self.user_login.user_token,\
                                                                        self.user_login.http_client.mac)
            self.cloud_sync_controller = cloud_sync_controller.CloudSyncController()
        else:
            self.logger.error("Can't retrieve user login or printer profiles. Exiting...")
            self.quit()

    @staticmethod
    def open_browser():
        if config.get_settings()['web_interface']['browser_opening_on_start']:
            import webbrowser
            import logging
            address = "http://%s:%d" % (config.LOCAL_IP, config.get_settings()['web_interface']['port'])
            logger = logging.getLogger()
            logging.basicConfig()
            logger.info("Opening browser tab %s" % address)
            webbrowser.open(address, 2, True)

    def wait_and_open_browser(self):
        start = time.time()
        self.logger.info("Waiting for timeout or login to open browser...")
        while time.time() < start + self.BROWSER_OPENING_DELAY and not self.user_login.user_token:
            time.sleep(0.01)
        self.logger.info("...done")
        App.open_browser()

    def init_interface(self):
        if config.get_settings()['web_interface']['enabled']:
            from web_interface import WebInterface
            self.web_interface = WebInterface(self)
            self.web_interface.start()
            self.logger.debug("Waiting for webserver to start...")
            while not self.web_interface.server:
                time.sleep(0.01)
                if self.stop_flag:
                    return
            self.logger.debug("...server is up and running.")

    def intercept_signal(self, signal_code, frame):
        self.logger.warning("SIGINT or SIGTERM received. Closing 3DPrinterOS Client version %s_%s" % \
                (version.version, version.build))
        self.stop_flag = True

    def toggle_virtual_printer(self):
        self.virtual_printer_enabled = not self.virtual_printer_enabled
        if not self.virtual_printer_enabled:
            for printer_interface in self.printer_interfaces:
                usb_info = getattr(printer_interface, 'usb_info', None)
                if usb_info and usb_info == self.VIRTUAL_PRINTER_USB_INFO:
                    printer_interface.close()
        settings = config.get_settings()
        settings['virtual_printer']['enabled'] = self.virtual_printer_enabled
        config.Config.instance().save_settings(settings)

    @log.log_exception
    def start_main_loop(self): #TODO cleanup from network and virtual code. main loop code should be short.
        self.last_time = 0
        while not self.stop_flag:
            self.updater.timer_check_for_updates()
            try:
                detected_printers = self.detector.get_printers_list() + self.network_printers_detector.get_printers_list()
            except Exception:
                self.logger.exception("Detector error: ")
            else:
                if self.virtual_printer_enabled:
                    detected_printers.append(self.VIRTUAL_PRINTER_USB_INFO)
                currently_connected_usb_info = [pi.usb_info for pi in self.printer_interfaces]
                for usb_info in detected_printers:
                    if usb_info not in currently_connected_usb_info:
                        pi = printer_interface.PrinterInterface(usb_info, self.user_login)
                        pi.start()
                        self.printer_interfaces.append(pi)
                for pi in self.printer_interfaces:
                    if pi.usb_info not in detected_printers:
                        if not pi.error_state == "error":
                            config.create_error_report(99, "Printer is no longer detected as USB device.",
                                                       pi.usb_info, self.logger, is_blocking=True)
                    if not pi.is_alive():
                        if not pi.stop_flag:
                            self.logger.warning("Printer interface of %s had crushed" % pi.usb_info)
                        self.logger.info('Removing %s from printer list' % pi.usb_info)
                        self.printer_interfaces.remove(pi)
            if not self.stop_flag:
                time.sleep(self.MAIN_LOOP_SLEEP)
            now = time.time()
            if self.last_time + self.IDLE_RECORD_TIMEOUT > now:
                self.last_time = now
                self.logger.debug(time.strftime("%d %b %Y %H:%M:%S", time.localtime()))
        return self.quit()

    def shutdown_web_interface(self):
        self.logger.debug("Waiting web interface server to shutdown...")
        try:
            self.web_interface.server.shutdown()
            self.web_interface.join()
        except:
            pass
        self.logger.debug("...done")

    def quit(self): # TODO reduce copy-paste
        self.logger.info("Starting exit sequence...")
        if hasattr(self, 'tray_controller'):
            state = 'Closing tray...'
            self.logger.info(state)
            self.closing_status.append(state)
            self.tray_controller.stop()
            self.closing_status[-1] += 'closed'
            time.sleep(0.5)
        if hasattr(self, 'camera_controller'):
            state = 'Closing camera...'
            self.logger.info(state)
            self.closing_status.append(state)
            self.camera_controller.stop_camera_process()
            self.closing_status[-1] += 'closed'
            time.sleep(0.5)
        if hasattr(self, 'cloud_sync_controller'):
            state = 'Closing CloudSync...'
            self.logger.info(state)
            self.closing_status.append(state)
            self.cloud_sync_controller.stop_cloud_sync_process()
            self.closing_status[-1] += 'closed'
            time.sleep(0.5)
        for pi in self.printer_interfaces:
            pi.close()
        for pi in self.printer_interfaces:
            printer_name = 'undefined printer'
            if pi.printer_profile and pi.printer_profile.get('name'):
                printer_name = pi.printer_profile['name']
            state = 'Closing ' + printer_name + '...'
            self.logger.info(state)
            self.closing_status.append(state)
            pi.join()
            self.closing_status[-1] += 'closed'
            time.sleep(0.5)
        time.sleep(0.2) #to reduce logging spam in next
        self.shutdown_web_interface()
        self.logger.info("...everything is correctly closed.")
        self.logger.info("Goodbye ;-)")
        for handler in self.logger.handlers:
            handler.flush()
        return self.restart_flag

if __name__ == '__main__':
    app = App()
    app.start_main_loop()


