#Copyright (c) 2015 3D Control Systems LTD

#3DPrinterOS client is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#3DPrinterOS client is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with 3DPrinterOS client.  If not, see <http://www.gnu.org/licenses/>.

# Author: Oleg Panasevych <panasevychol@gmail.com>, Vladimir Avdeev <another.vic@yandex.ru>

import re
import sys
import time
import json
import uuid
import httplib
import logging
from subprocess import Popen, PIPE

import config
import version
import platforms
from checker_waiter import CheckerWaiter

# machine id is mac address, but on RPi we use machine serial
def get_serial_number():
    if sys.platform.startswith('linux'):
        call = Popen(['cat', '/proc/cpuinfo'], stdout=PIPE, stderr=PIPE)
        stdout, stderr = call.communicate()
        stdout = stdout.replace('\t', '').split('\n')
        for item in stdout:
            if 'Serial: ' in item:
                return item.replace('Serial: ', '')

class HTTPClient:

    URL = config.get_settings()['URL']
    HTTPS_MODE = config.get_settings()['HTTPS']
    BASE_TIMEOUT = 6
    MAX_FAILS = 2
    MAX_TIMEOUT = BASE_TIMEOUT * 3

    streamer_prefix = "/streamerapi"
    user_login_path = streamer_prefix + "/user_login"
    printer_login_path = streamer_prefix + "/printer_login"
    command_path = streamer_prefix + "/command"

    token_send_logs_path = streamer_prefix + "/sendLogs"
    camera_path = streamer_prefix + "/camera" #json['image': base64_image ]
    cloudsync_path = "/autoupload"

    domain_path_re = re.compile("https?:\/\/(.+)(\/.*)")

    def __init__(self, parent, keep_connection_flag = False, debug = False):
        self.logger = logging.getLogger(__name__)
        self.parent = parent
        self.parent_usb_info = getattr(parent, 'usb_info', None)
        if debug:
            self.logger.setLevel('DEBUG')
        else:
            self.logger.setLevel('INFO')
        self.keep_connection_flag = keep_connection_flag
        self.timeout = self.BASE_TIMEOUT
        self.connection = self.connect()
        self.fails = 0

    def get_mac_add_or_serial_number(self):
        id = get_serial_number()
        if not id:
            id = self.get_mac_for_current_ip()
        if not id:
            time.sleep(1)
            id = self.get_mac_for_current_ip()
        if not id:
            self.logger.error("Can't get mac address! Using uuid.getnode().")
            id = hex(uuid.getnode())
        return id

    def format_mac_addr(self, mac):
        mac = mac.replace(":", "")
        mac = mac.replace("-", "")
        mac = mac.lower()
        return "0x" + mac + "L"

    def get_mac_for_current_ip(self):
        if platforms.PLATFORM in ("rpi", 'linux', "mac"):
            process = Popen(['ifconfig'], stdout=PIPE)
            stdout, _ = process.communicate()
            interfaces = stdout.split("flags=") #will get name of interface wrong, but we don't need name
            for interface in interfaces:
                if "inet " + self.local_ip in interface:
                    search = re.search("ether\s([0-9a-f\:]+)", interface)
                    if search:
                        return self.format_mac_addr(search.group(1))
        else:
            process = Popen(['ipconfig' , "/all"], shell=True, stdout=PIPE)
            stdout, _ = process.communicate()
            interfaces = stdout.split("\r\n\r\n")
            for interface in interfaces:
                search = re.search("IP.*:\s(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})", interface)
                if search:
                    ip = search.group(1)
                    if ip == self.local_ip:
                        search = re.search("[A-F0-9]{2}\-[A-F0-9]{2}\-[A-F0-9]{2}\-[A-F0-9]{2}\-[A-F0-9]{2}\-[A-F0-9]{2}", interface)
                        if search:
                            return self.format_mac_addr(search.group(0))

    def connect(self):
        self.logger.debug("{ Connecting...")
        fails = 0
        while not self.parent.stop_flag:
            try:
                if self.HTTPS_MODE:
                    connection = httplib.HTTPSConnection(self.URL, port = 443, timeout = self.timeout)
                else:
                    connection = httplib.HTTPConnection(self.URL, port = 80, timeout = self.timeout)
                connection.connect()
                self.local_ip = connection.sock.getsockname()[0]
                self.mac = self.get_mac_add_or_serial_number()
            except Exception as e:
                config.create_error_report(5, "Error during HTTP connection: " + str(e),
                                           self.parent_usb_info, self.logger, is_blocking=False)
                self.logger.debug("...failed }")
                self.logger.warning("Warning: connection to %s failed." % self.URL)
                if self.timeout < self.MAX_TIMEOUT:
                    self.timeout += self.BASE_TIMEOUT
                time.sleep(1)
                fails += 1
                if fails > self.MAX_FAILS:
                    return
            else:
                self.logger.debug("...success }")
                self.logger.info("Connecting from: %s\t%s" % (self.local_ip, self.mac))
                return connection

    def load_json(self, jdata):
        try:
            data = json.loads(jdata)
        except ValueError:
            config.create_error_report(2, "Received data is not valid json: "
                                       + jdata, self.parent_usb_info, self.logger, is_blocking=False)
        else:
            if data and type(data) == dict:
                return data
            else:
                config.create_error_report(3, "Error parsing http message, data should be not empty dictionary: " +
                                           str(data), self.parent_usb_info, self.logger, is_blocking=False)

    def request(self, method, connection, path, payload, headers=None):
        self.logger.debug("{ Requesting...")
        if headers is None:
            headers = {"Content-Type": "application/json", "Content-Length": str(len(payload))}
            if self.keep_connection_flag:
                headers["Connection"] = "keep-alive"
        try:
            connection.request(method, path, payload, headers)
            resp = connection.getresponse()
        except Exception as e:
            config.create_error_report(6, "Error during HTTP request:" + str(e),
                                       self.parent_usb_info, self.logger, is_blocking=False)
        else:
            #self.logger.debug("Request status: %s %s" % (resp.status, resp.reason))
            try:
                received = resp.read()
            except Exception as e:
                config.create_error_report(7, "Error reading response: " + str(e),
                                           self.parent_usb_info, self.logger, is_blocking=False)
            else:
                if resp.status == httplib.OK and resp.reason == "OK":
                    self.logger.debug("...success }")
                    return received
                else:
                    config.create_error_report(8, "Error: server responded with non 200 OK\nFull response: %s" %
                                               received, self.parent_usb_info, self.logger, is_blocking=False)
        self.logger.debug("...failed }")
        self.logger.warning("Warning: HTTP request failed!")

    def pack_and_send(self, target, *payloads, **kwargs_payloads):
        path, packed_message = self.pack(target, *payloads, **kwargs_payloads)
        return self.send(path, packed_message)

    def send(self, path, data):
        fails = 0
        while not self.parent.stop_flag:
            if not self.connection:
                self.connection = self.connect()
            if self.connection:
                answer = self.request("POST", self.connection, path, data)
                if not self.keep_connection_flag:
                    self.close()
                if answer:
                    self.fails = 0
                    return self.load_json(answer)
                elif self.connection:
                    self.close()
            time.sleep(1)
            fails += 1
            if fails > self.MAX_FAILS:
                return

    def pack(self, target, *args, **kwargs):
        if target == 'user_login':
            data = { 'login': {'user': args[0], 'password': args[1]},
                     'platform': platforms.PLATFORM, 'host_mac': self.mac,
                     'local_ip': self.local_ip, 'version': version.version, 'build': version.build }
            if kwargs.get('auth_token'):
                data['login']['disposable_token'] = kwargs['auth_token']
            path = self.user_login_path
            #self.logger.debug(data)
        elif target == 'printer_login':
            data = { 'user_token': args[0], 'printer': args[1], 'version': version.version, 'build': version.build,
                     'data_time': time.ctime(), 'camera': config.get_app().camera_controller.get_current_camera_name() }
            path = self.printer_login_path
        elif target == 'command':
            data = { 'printer_token': args[0], 'report': args[1], 'command_ack': args[2] }
            if not data['command_ack']:
                data.pop('command_ack')
            path = self.command_path
        elif target == 'camera':
            data = { 'user_token': args[0], 'camera_number': args[1], 'camera_name': args[2],
                     'file_data': args[3], 'host_mac': self.mac}
            path = self.camera_path
        elif target == 'cloudsync':
            data = { 'user_token': args[0], 'file_data': args[1]}
            path = self.cloudsync_path
        else:
            config.create_error_report(4, 'No such target for packaging: ' + target, None,
                                       self.logger, is_blocking=False)
            data, path = None, None
        for key, value in kwargs.items():
            data[key] = value
        return path, json.dumps(data)

    def close(self):
        if self.connection:
            self.connection.close()
            self.connect = None


class NetworkConnectionChecker(CheckerWaiter):

    NAME_FOR_LOGGING = 'network connection check'

    def __init__(self, app):
        self.app = app
        CheckerWaiter.__init__(self, app, self.check_connection)

    def check_connection(self):
        http_client = HTTPClient(self.app)
        return bool(http_client.connection)

if __name__ == '__main__':
    http_client = HTTPClient()